#pragma once

#include "ap_pal.h"

#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

#ifndef LIFEN_TODO
#include "ApPkiClient.h"
#endif

#include "DMClientImpl.h"

#ifndef PLATFORM_UNIX
#include "HttpRequest.h"
#else
#include "httprequest_linux.h"
#endif

#include <wincrypt.h>


namespace dmclient
{
    /// <summary>
    /// Connects to the DMGateway using the DMGateway's application protocol that uses HTTP.
    /// Calls to the APSDK DMClient are delegated to methods here, if the cluster / environment / service
    /// settings indicate that DMClient needs to communicate with DMGateway and not ApManagementProxy.
    /// </summary>
    class DMClientHttpImpl : public DMClientImpl
    {
        class HttpEndpointContainer;

    public:

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="clusterSettings">The settings for DMClient configured at the cluster hierarchy (in shared.ini).
        /// Amongst the various settings that can be configured there, the primary ones that are used here are the
        /// DMGateway endpoint URL, port number, and whether HTTPS should be used.</param>
        /// <param name="environmentSettings">The settings for DMClient configured in environmnent.ini.</param>
        /// <param name="applicationSettings">The settings for DMClient configured in the application config.ini.</param>
        DMClientHttpImpl(
            _In_ std::shared_ptr<DMClientSettings> clusterSettings,
            _In_ std::shared_ptr<DMClientSettings> environmentSettings,
            _In_ std::shared_ptr<DMClientSettings> applicationSettings);

#ifdef PLATFORM_UNIX
        /// <summary>
        /// Destructor.
        /// </summary>
        ~DMClientHttpImpl();

        void Close();
#endif
        /// <summary>
        /// Invoked by the base DMClientImpl class, for any implementation specific initialization.
        /// Here, we create the HttpEndpoint objects corresponding to the DMGateway endpoints.
        /// </summary>
        virtual bool OnInit() override;

        /// <summary>
        /// Sends a command to the DM Gateway.
        /// </summary>
        /// <param name="thisPtr">The shared_ptr object that is storing the pointer to the current DMClientHttpImpl instance.</param>
        /// <param name="req">Request encapsulating the command to be sent plus any additional payload.</param>
        /// <param name="resp">Response from the DM Gateway.</param>
        /// <param name="ctx">Command context. Used to indicate whether the command should be executed asynchronously or not.</param>
        /// <returns>true if the command was successfully sent. Else false.</returns>
        virtual bool Send(_In_ std::shared_ptr<DMClientImpl> thisPtr, _In_ DMRequest* req, _In_ DMResponse *resp, _In_ void *ctx) override;

    private:

        /// <summary>
        /// Used as the key comparator in std::map for performing case-insensitive comparisons,
        /// when the key is std::string
        /// </summary>
        struct LessI
        {
            bool operator()(std::string left, std::string right) const
            {
                return (_stricmp(left.c_str(), right.c_str()) < 0);
            }
        };

        /// <summary>
        /// This class implements HttpRequest::CallbackHandler. Instances of this class
        /// are used to get notifications once the asynchronous HTTP call completes.
        /// </summary>
        class CallbackHandler : public HttpRequest::CallbackHandler
        {
        public:
            /// <summary>
            /// Constructor.
            /// </summary>
            CallbackHandler();

            /// <summary>
            /// Destructor.
            /// </summary>
            ~CallbackHandler();

            /// <summary>
            /// Invoked by the HTTP client when an error occurs while processing
            /// the asynchronous call.
            /// </summary>
            /// <param name="errorCode">The error code corresponding to the error. This
            /// will be a Windows error code, which can be looked up on MSDN.</param>
            virtual void OnError(unsigned long errorCode) override;

            /// <summary>
            /// Invoked by the HTTP client when a response has been received from the server.
            /// </summary>
            /// <param name="response">The object containing the response.</param>
            virtual void OnResponseReceived(_In_ HttpResponse* response) override;

            /// <summary>
            /// The value of the serverCookie when the asynchronous call was made. This
            /// value is provided by the DMClientHttpImpl parent instance.
            /// </summary>
            int64_t m_serverCookie;

            /// <summary>
            /// The HTTP server to which we're connecting.
            /// </summary>
            std::string m_server;

            /// <summary>
            /// Indicates whether the HTTP server to which we're connecting is vip or dip.
            /// True if vip, false otherwise.
            /// </summary>
            bool isServerDmGatewayVip;

            /// <summary>
            /// The parent DMClientHttpImpl instance on which the asynchronous call was made.
            /// We use this to get access to the callback function provided by the end-user, so that
            /// we can invoke it once the asynchronous call has completed.
            /// </summary>
            std::shared_ptr<DMClientHttpImpl> m_dmClientHttpImpl;

            /// <summary>
            /// The DMResponse object provided by the end-user, which needs to be populated with the
            /// response from the DMGateway once the asynchronous call has completed.
            /// </summary>
            DMResponse* m_dmResponse;

            /// <summary>
            /// The context provided by the end-user when invoking the Send() method asynchronously
            /// on the DM Client. We need to pass this to the callback function when the call completes.
            /// </summary>
            void* m_asyncContext;

            /// <summary>
            /// The trace identifier for the request as provided by the user, or the originally generated one.
            /// On subsequent retries, this trace identifier will be appended with the retry number.
            /// </summary>
            std::string m_originalTraceId;

            /// <summary>
            /// The trace identifier for the request. This is m_originalTraceId plus the retry number, in case
            /// of retries.
            /// </summary>
            std::string m_traceId;

            /// <summary>
            /// The command being executed by the user.
            /// </summary>
            std::string m_cmd;

            /// <summary>
            /// Gets the value of the readOnly parameter specified by the user when invoking DMClient.Send().
            /// </summary>
            bool m_cmdReadOnly;

            /// <summary>
            /// The HTTP Request object on which the asynchronous call was made.
            /// </summary>
            std::shared_ptr<HttpRequest> m_httpRequest;

            /// <summary>
            /// The request headers that need to be sent to DMGateway corresponding to this request.
            /// These headers are named "static" since they don't change with each subsequent retry (e.g.,
            /// these headers don't contain the trace id).
            /// </summary>
#ifndef PLATFORM_UNIX
            std::string m_staticRequestHeaders;
#else
            std::map<std::string, std::string> m_staticRequestHeaders;
#endif

            /// <summary>
            /// This event, if non-null, will be signalled once the call completes. This is used only
            /// when the user is executing the request synchronously (for async calls, callbacks are used).
            /// </summary>
            HANDLE m_callCompletedEvent;

            /// <summary>
            /// Gets the status of the command execution for the most recent retry.
            /// </summary>
            bool m_success;

            /// <summary>
            /// Gets the count of the current retry. A value of 0 implies this is the first execution of
            /// the request. A value of 'n' implies the 'n'th retry (and thus a total of 'n'+1 executions
            /// of the request will have been sent to DMGateway).
            /// </summary>
            int m_retryCount;

            /// <summary>
            /// This is initialized to the time when we made the first call to DMGateway to execute this
            /// request. This is used to ensure that we don't keep retrying failed calls if the timeout
            /// specified by the caller has already elapsed.
            /// </summary>
            LARGE_INTEGER m_startTime;

            /// <summary>
            /// This is initialized to the time when we made the last call to DMGateway to execute this
            /// request. This is used to calculate response latency for each request.
            /// </summary>
            LARGE_INTEGER m_lastRequestSentTime;

        private:

            /// <summary>
            /// Helper method to either retry, or invoke the user callback to indicate that the call has completed
            /// (if cannot retry), or in the case of a synchronous call, to indicate that the call has completed (if
            /// cannot retry).
            /// The DMResponse object (provided by the user) is also populated with the response details.
            /// </summary>
            /// <param name="errorCode">The error code corresponding to the error. This
            /// will be a Windows/WinHttp error code, which can be looked up on MSDN.</param>
            /// <param name="httpResponse">The HTTP response obtained from DMGateway, if we
            /// received one.</param>
            void RetryOrNotifyUserOnCallComplete(unsigned long errorCode, _In_ HttpResponse* httpResponse);
        };

        /// <summary>
        /// The version of the DMClient->DMGateway communication protocol.
        /// </summary>
        enum class Version
        {
            InitialVersion = 1,
            UserNameEnvironmentVariable = 2,
            UseDmGateway2 = 3,
            // Insert new version here.
            NextVersion
        };

        /// <summary>
        /// This method is used to determine whether the command can be retried, based on the
        /// command characteristics / response received.
        /// </summary>
        /// <param name="callbackHandler">The callback handler instance associated with the request.</param>
        /// <param name="errorCode">The WinHTTP error code.</param>
        /// <param name="httpResponse">The response received from DMGateway, if any.</param>
        /// <returns>Whether the command can be retried or not.</returns>
        bool CanRetry(
            _In_ const CallbackHandler* callbackHandler,
            const unsigned long errorCode,
            _In_opt_ const HttpResponse* httpResponse);

        /// <summary>
        /// Dispatches the HTTP request for asynchronous execution.
        /// If the request was processed earlier and had failed (based on previousErrorCode and previousHttpResponse),
        /// this request is dispatched again only if it can be retried.
        /// If dispatching fails (WinHttp returns an error), then this is retried until it cannot be retried any more
        /// (based on retry count, e.g.).
        /// </summary>
        /// <param name="callbackHandler">The callback handler instance associated with the request.</param>
        /// <param name="previousErrorCode">The error code from the previous execution of this request, if any (or 0 if this
        /// is the first invocation).</param>
        /// <param name="previousHttpResponse">The HTTP response obtained from the previous execution of this request, if any (or NULL if this
        /// is the first invocation).</param>
        /// <returns>The error code / HTTP response returned from dispatching the call.
        /// Since when the HTTP call is dispatched we don't yet have an HTTP response - the HTTP Response returned will always be NULL if the
        /// call was dispatched. If the call wasn't dispatched - e.g., we didn't retry the request because one of the criteria wasn't met,
        /// then this will be the same as what was passed in.
        /// Similarly for the error code in the return pair - it is either the return code due to dispatching the request, or what was passed
        /// in if the call wasn't dispatched.</returns>
        std::pair<unsigned long, HttpResponse*> Send(_In_ CallbackHandler* callbackHandler, unsigned long previousErrorCode, _In_ HttpResponse* previousHttpResponse);

        /// <summary>
        /// Returns a boolean that indicates whether the DMClient when connecting to the DMGateway should use the
        /// machine certificate for authentication, or whether it should use the default certificate returned by the AP PKI APIs
        /// (which would return the machine certificate if the credentials under which the service is running are machine credentials
        /// such as LOCAL MACHINE; or would return the user-certificate associated with the user-account under which the service
        /// is running).
        /// This is based on settings configured at the cluster level (via shared.ini),
        /// environment level (via environment.ini) as well as the service level (in the default config used to
        /// initialize the AP runtime).
        /// Service config takes precedence over environment config, which takes precedence over cluster config.
        /// Unless, the cluster config has indicated that settings at the service/environment level need to be ignored,
        /// in which case this only depends on the settings configured at cluster level (via shared.ini).
        /// </summary>
        /// <param name="clusterSettings">The settings for DMClient configured at the cluster hierarchy (in shared.ini).</param>
        /// <param name="environmentSettings">The settings for DMClient configured in environmnent.ini.</param>
        /// <param name="applicationSettings">The settings for DMClient configured in the application config.ini.</param>
        /// <returns>A boolean indicating whether only machine certificates should be used, or use the certificate associated
        /// with the service credentials.</returns>
        static bool GetOnlyUseMachineCertificate(
            _In_ std::shared_ptr<DMClientSettings> clusterSettings,
            _In_ std::shared_ptr<DMClientSettings> environmentSettings,
            _In_ std::shared_ptr<DMClientSettings> applicationSettings);

        /// <summary>
        /// Gets settings that can be configured in the configuration file.
        /// Settings can be configured at the cluster level (via shared.ini),
        /// environment level (via environment.ini) as well as the service level (in the default config used to
        /// initialize the AP runtime).
        /// Service config takes precedence over environment config, which takes precedence over cluster config.
        /// </summary>
        /// <param name="clusterSettings">The settings for DMClient configured at the cluster hierarchy (in shared.ini).</param>
        /// <param name="environmentSettings">The settings for DMClient configured in environmnent.ini.</param>
        /// <param name="applicationSettings">The settings for DMClient configured in the application config.ini.</param>
        /// <param name="maxConnectionsPerServer">The maximum number of connections that should be opened to a single DMGateway endpoint.</param>
        /// <param name="numRetries">The number of times a request can be retried, after the initial execution fails.</param>
        /// <param name="ignoreTimeoutDuringRetry">Whether the timeout should be ignored during a retry.</param>
        static void GetSettings(
            _In_ std::shared_ptr<DMClientSettings> clusterSettings,
            _In_ std::shared_ptr<DMClientSettings> environmentSettings,
            _In_ std::shared_ptr<DMClientSettings> applicationSettings,
            _Out_ unsigned long& connectTimeoutMilliseconds,
            _Out_ unsigned long& resolveTimeoutMilliseconds,
            _Out_ unsigned long& maxConnectionsPerServer,
            _Out_ int& numRetries,
            _Out_ bool& ignoreTimeoutDuringRetry);

        /// <summary>
        /// Gets dmgateway server validator function that will be used for server authentication against given acl.
        /// </summary>
        /// <param name="serverCertificateAcl">server certificate acl against which server should be authenticated.</param>
#ifndef PLATFORM_UNIX
        static std::function<bool(PCERT_CONTEXT)> GetDmGatewayServerCertificateValidator(_In_ const std::string& serverCertificateAcl);
#else
        static std::function<bool(bool, boost::asio::ssl::verify_context&)> GetDmGatewayServerCertificateValidator(_In_ const std::string& serverCertificateAcl);
#endif

        /// <summary>
        /// Determines if a failure is due to connecting to a specific DMGateway endpoint (either connection
        /// failed, or the DMGateway endpoint didn't process the request, etc).
        /// </summary>
        /// <param name="errorCode">The WinHttp error code.</param>
        /// <param name="httpResponse">The HTTP response received from the DMGateway endpoint, if any.</param>
        /// <returns>Whether the error is due to connection issues or DMGateway endpoint issues (and not
        /// due to a problem with the request being executed).</returns>
        static bool IsEndpointError(_In_ const unsigned long errorCode, _In_ const HttpResponse* httpResponse);

        /// <summary>
        /// Returns whether DMGateway vip should be used to send requests based on http endpoint container settings and
        /// if this is a first time request.
        /// </summary>
        /// <param name="isFirstTimeRequest">Indicates if this the first time request.</param>
        bool shouldUseDmGatewayVipEndPoint(
            _In_ const bool isFirstTimeRequest);

        /// <summary>
        /// Creates a new http endpoint corresponding to the input server.
        /// </summary>
        /// <param name="server">server for which http endpoint is to be created.</param>
        /// <param name="dmGatewayServerCertificateValidator">server certificate validator function to authenticate server certificate</param>
#ifndef PLATFORM_UNIX
        std::shared_ptr<HttpEndpoint> CreateNewHttpEndpoint(
            _In_ const std::string& server,
            _In_ const std::function<bool(PCERT_CONTEXT)>& dmGatewayServerCertificateValidator);
#else
        std::shared_ptr<HttpEndpoint> CreateNewHttpEndpoint(
            _In_ const std::string& server,
            _In_ const std::function<bool(bool, boost::asio::ssl::verify_context&)>& dmGatewayServerCertificateValidator);
#endif

        /// <summary>
        /// Adds the input server to the given http end point container.
        /// </summary>
        /// <param name="httpEndpointContainer">HttpEndPoint container to which the input server is to be added.</param>
        /// <param name="server">Server to be added to http end point container</param>
        /// <param name="dmGatewayServerCertificateValidator">server certificate validator function to authenticate server certificate</param>
#ifndef PLATFORM_UNIX
        bool AddServerToHttpEndPointContainer(
            _Inout_ std::shared_ptr<HttpEndpointContainer> httpEndpointContainer,
            _In_ const std::string& server,
            _In_ const std::function<bool(PCERT_CONTEXT)>& dmGatewayServerCertificateValidator);
#else
        bool AddServerToHttpEndPointContainer(
            _Inout_ std::shared_ptr<HttpEndpointContainer> httpEndpointContainer,
            _In_ const std::string& server,
            _In_ const std::function<bool(bool, boost::asio::ssl::verify_context&)>& dmGatewayServerCertificateValidator);
#endif

        /// <summary>
        /// The data received from the DM Gateway, within the HttpResponse object, is copied over
        /// to the DMResponse object, which the end-user sees.
        /// </summary>
        /// <param name="traceId">The trace identifier associated with the request.</param>
        /// <param name="errorCode">The error code for the error that occurred (if any) during communication
        /// with the DMGateway HTTP endpoint. The code will either be 0 (for no error) or a Windows error code.</param>
        /// <param name="httpResponse">The HTTP response received from the HTTP server (the DMGateway endpoint).</param>
        /// <param name="response">The DMResponse object provided by the end-user.</param>
        /// <param name="serverCookie">The server cookie corresponding to the DMGateway endpoint which was used to send
        /// this request.</param>
        /// <param name="server">The HTTP server to which we connected and got the error / response from.</param>
        void PopulateDMResponse(_In_z_ const char* traceId, unsigned long errorCode, _In_ HttpResponse* httpResponse, _In_ DMResponse* response, int64_t serverCookie, _In_z_ const char* server);

        /// <summary>
        /// Gets the trace identifier to use for the request. The user-provided one is used, and if none
        /// is provided, then we generate one.
        /// </summary>
        /// <param name="req">The request.</param>
        /// <param name="traceId">This is populated with the trace identifier to use.</param>
        void GetTraceId(_In_ const DMRequest* req, _Out_ std::string& traceId);

        /// <summary>
        /// Creates the HTTP Request header that needs to be sent to the DMGateway. The header is created
        /// from the values in the DMRequest (e.g., Command) as well as some caller metadata (e.g., caller's
        /// environment name).
        /// The headers constructed here do not change even across retries.
        /// </summary>
        /// <param name="req">The DMRequest object containing the request data from which the HTTP header has to be
        /// constructed.</param>
        /// <param name="header">The request header is created and populated into this parameter.</param>
#ifndef PLATFORM_UNIX
        void GenerateStaticRequestHeader(_In_ const DMRequest* req, _Out_ std::string& header);
#else
        void GenerateStaticRequestHeader(_In_ const DMRequest* req, _Out_ std::map<std::string, std::string>& header);
#endif

        /// <summary>
        /// Creates the HTTP Request header that needs to be sent to the DMGateway.
        /// The static request headers are combined with the dynamic data (that change across retries).
        /// </summary>
        /// <param name="staticRequestHeaders">The request headers to which additional headers will be added.</param>
        /// <param name="traceId">The trace identifier corresponding to this request.</param>
        /// <param name="header">The request header is created and populated into this parameter.</param>
#ifndef PLATFORM_UNIX
        void GenerateRequestHeader(_In_ const std::string& staticRequestHeaders, _In_ const std::string& traceId, _Out_ std::string& header);
#else
        void GenerateRequestHeader(_In_ const std::map<std::string, std::string>& staticRequestHeaders, _In_ const std::string& traceId, _Out_ std::map<std::string, std::string>& header);
#endif

        /// <summary>
        /// Converts the HTTP status code (received from DMGateway) to a DMResponse status code.
        /// </summary>
        /// <param name="httpStatusCode">The HTTP status code.</param>
        /// <param name="responseHeaders">The response HTTP headers.</param>
        DMResponse::StatusCode ConvertHttpStatusCodeToDMResponseStatusCode(unsigned long httpStatusCode, const std::map<std::string, std::string>& responseHeaders);

        /// <summary>
        /// Converts the Windows error code (received from WinHTTP) to a DMResponse status code.
        /// </summary>
        DMResponse::StatusCode ConvertWindowsErrorCodeToDMResponseStatusCode(unsigned long errorCode);

        /// <summary>
        /// Maps the string representation of DMResponse::StatusCode to the actual enum value.
        /// Returns the mapping.
        /// </summary>
        static std::map<std::string, DMResponse::StatusCode, LessI> CreateStatusCodeMap();

#ifdef PLATFORM_UNIX
        /// <summary>
        /// The Boost io_context object provides the core I/O functionality for users of the asynchronous I/O objects
        /// </summary>
        boost::asio::io_context m_io_context;

        /// <summary>
        /// The Boost IO context worker.
        /// The work class is used to inform the io_context when work starts and finishes.
        /// This ensures that the io_context object's run() function will not exit while work is underway
        /// </summary>
        std::unique_ptr<boost::asio::io_context::work> m_work;

        /// <summary>
        /// The thread on which Boost IO context worker runs.
        /// </summary>
        std::unique_ptr<std::thread> m_io_thread;
#endif

        /// <summary>
        /// The HTTP resource URL at which the DMGateway is listening for incoming requests.
        /// In the URI, this is the portion that comes after the hostname.
        /// </summary>
        std::string m_resourceUrl;

        /// <summary>
        /// The port number to which DMClientHttpImpl will connect. DMGateway will be listening at that port.
        /// </summary>
        unsigned short m_port;

        /// <summary>
        /// Indicates whether the DMClientHttpImpl makes a HTTPS connection to the DMGateway, or a HTTP connection.
        /// This value is read from the cluster configuration.
        /// </summary>
        bool m_useHttps;

        /// <summary>
        /// Indicates whether the Machine AP PKI certificate should be used, or the default certificate (based on
        /// the user context).
        /// </summary>
        bool m_useMachineCertificate;

        /// <summary>
        /// The AP PKI ACL that will be used to verify the server SSL certificate (in case of HTTPS connections).
        /// </summary>
        std::string m_serverCertificateAcl;

        /// <summary>
        /// The maximum number of connections to open per DMGateway endpoint.
        /// </summary>
        unsigned long m_maxConnectionsPerServer;

        /// <summary>
        /// The timeout for connecting to the DMGateway endpoint.
        /// Passed on to WinHttp.
        /// </summary>
        unsigned long m_connectTimeoutMilliseconds;

        /// <summary>
        /// The timeout for resolving the host name of the DMGateway server to the IP address.
        /// Passed on to WinHttp.
        /// </summary>
        unsigned long m_resolveTimeoutMilliseconds;

        /// <summary>
        /// The number of times to retry, after the initial request execution has failed.
        /// </summary>
        int m_numRetries;

        /// <summary>
        /// Whether we should ignore the timeout during retries.
        /// </summary>
        bool m_ignoreTimeoutDuringRetry;

        /// <summary>
        /// The current version of the DMClient->DMGateway communication protocol.
        /// </summary>
        static const Version c_versionNumber;

        /// <summary>
        /// The delimiter between two HTTP headers, which is \r\n.
        /// </summary>
        static const char* c_httpHeadersDelimiter;

        /// <summary>
        /// The HTTP header in the response which contains the DM Response Code. If this is present, its value should
        /// be used as the DMResponse code, instead of determining the code from the HTTP status code.
        /// </summary>
        static const char* c_dmResponseCodeHeader;

        /// <summary>
        /// The HTTP header in the response, which indicates if the request should not be retried.
        /// </summary>
        static const char* c_noRetryHeader;

        /// <summary>
        /// Maps the string representation of the status code, to the status code.
        /// The string representation of the status code is what will be sent to DMClient
        /// by DMGateway.
        /// </summary>
        static std::map<std::string, DMResponse::StatusCode, LessI> s_statusCodeMap;

        /// <summary>
        /// This is the DMGateway vip endpoint.
        /// If DMGateway vip is not enabled, this has null value.
        /// </summary>
        std::shared_ptr<HttpEndpoint> m_dmGatewayVipEndpoint;

        /// <summary>
        /// The HttpEndpointContainer instance which this DMClientHttpImpl instance will use
        /// for connecting to DMGateway. This has null value if the instance is configured to use only vip for sending requests.
        /// </summary>
        std::shared_ptr<HttpEndpointContainer> m_httpEndpointContainer;

        /// <summary>
        /// This mutex synchronizes access to s_httpEndpointContainerMap.
        /// </summary>
        static std::mutex s_httpEndpointContainerMapMutex;

        /// <summary>
        /// This static map is used to store HttpEndpointContainer instances, each instance representing a
        /// set of DMGateway endpoints.
        /// The Key for the map is a string that contains all the data that was used to initialize DMClientHttpImpl.
        /// It consists of the server names, the port number, whether https should be used, the timeouts, and the
        /// ACL used to validate the server certificate.
        /// NOTE: entries once added to this map are never removed. The assumption is that in a single process,
        /// there is only a limited set of different combinations of inputs used to initialize DMClient.
        /// </summary>
        static std::map<std::string, std::shared_ptr<HttpEndpointContainer>, LessI> s_httpEndpointContainerMap;

        /// <summary>
        /// This class contains the HTTP endpoints to which DMClientHttpImpl instances will connect to. It also
        /// maintains the serverCookie that keeps track of which HTTP endpoint is currently being used.
        /// </summary>
        class HttpEndpointContainer
        {
        public:

            /// <summary>
            /// Constructor.
            /// </summary>
            HttpEndpointContainer();

            /// <summary>
            /// This is the list of DMGateway endpoints. An instance of DMClientHttpImpl will pick up one of these
            /// at random, and send all calls only to it. Whenever a communication error occurs (e.g., transport connection
            /// broken), or an internal server error occurs, then the DMClientHttpImpl will move on to the next endpoint
            /// in the list (or come back to the first one in the list) and then cycle through all again.
            /// We won't move to another endpoint if an application error has occurred (e.g., BadRequest, CommandNotFound, etc).
            /// </summary>
            std::vector<std::shared_ptr<HttpEndpoint>> m_httpEndpoints;

            /// <summary>
            /// The AP PKI ACL that will be used to verify the server SSL certificate (in case of HTTPS connections).
            /// </summary>
            std::string m_serverCertificateAcl;

            /// <summary>
            /// This cookie starts from 0 every time an instance of DMClientHttpImpl is created, and one endpoint
            /// from the available DMGateway endpoints is picked at random. Then, everytime we need to pick
            /// another endpoint (say, the existing endpoint is unreachable (e.g., DMGateway machine/service has
            /// gone down)), we move to the next endpoint in our list, by incrementing the value of this cookie.
            /// This cookie is used so that if multiple threads are hitting errors connecting to the same DMGateway
            /// endpoint, only one of them is able to increase cookie (to move to the next endpoint), and the other
            /// threads see that their cookie value doesn't match and thus don't increment the cookie (else we would
            /// end up "skipping" over some DMGateway endpoints in our list, and probably even land up at the non-functioning
            /// one again very soon.
            /// The cookie value is incremented from within the "UseNextHttpEndpoint" method.
            /// </summary>
            volatile int64_t m_serverCookie;

            /// <summary>
            /// From the available DM Gateway endpoints to which the DMClient can connect, one of the endpoints
            /// is chosen at random as the "start" DM Gateway endpoint. All DMClient calls are then sent only to that
            /// endpoint. Whenever an error occurs which requires the DMClient to now connect to some other DMGateway
            /// endpoint, this method should be invoked.
            /// This internally updates a "serverCookie" value which the DMHttpClientImpl maintains.
            /// This internal cookie is used for this scenario - assume a number of simultaneous calls are made asynchronously
            /// using DMClient to DMGateway. These all hit the same DMGateway endpoint. Now assume that the endpoint goes down,
            /// due to which all these connections fail. If all these calls now invoke UseNextHttpEndpoint() to internally
            /// update the state so that the next set of calls go to another DMGateway endpoint, we may end up internally "skipping"
            /// multiple DMGateway endpoints. Thus, each caller when invoking UseNextHttpEndpoint provides the cookie value
            /// at the time it was given an HttpEndpoint object. Then, UseNextHttpEndpoint will only indeed "use the next endpoint"
            /// only if the cookie values match. If the cookie values don't match, it means that another caller has already
            /// asked DMClientHttpImpl to use another HTTP endpoint, and that new endpoint may be working fine, and so we don't
            /// need to further switch to another HTTP endpoint.
            /// Any idle connections maintained by WinHTTP to the old HTTP endpoint will get closed after WinHTTP's internal timeout.
            /// </summary>
            /// <param name="serverCookie">The value of the cookie which the caller has; given to the caller by DMClientHttpImpl
            /// when the caller was provided an HttpEndpoint instance to connect to. The DMClientHttpImpl will internally only cycle
            /// to the next HttpEndpoint if the cookie value here matches what it has internally.</param>
            void UseNextHttpEndpoint(int64_t serverCookie);

            /// <summary>
            /// Gets the HttpEndpoint object corresponding to the DMGateway endpoint that the DMClient
            /// needs to connect to. Also gives out the cookie value corresponding to the current
            /// endpoint that is in rotation.
            /// </summary>
            /// <param name="serverCookie">This parameter is updated with the value of the cookie corresponding to the current
            /// endpoint that is in use.</param>
            std::shared_ptr<HttpEndpoint> GetHttpEndpoint(_Out_ int64_t* serverCookie);
        };
    };
};

