#pragma once

#include "ap_pal.h"

#include <set>

#include "basic_types.h"
#include "logging.h"
#include "DMClient.h"

namespace dmclient
{
    /// <summary>
    /// This represents a base class that is used internally by DMClient.
    /// Different implementations of this class can then do different things
    /// in order to process the command provided to DMClient.
    /// </summary>
    class DMClientImpl
    {
        friend class apsdk::DMClient;

    public:
        DMClientImpl() :
            m_callback(nullptr),
            m_keepAlive(false)
        {
            PopulateCallerMetadata();
        }

        virtual ~DMClientImpl() { }

        /// <summary>
        /// Initializes the DMClientImpl instance. This method must be called before this object is used
        /// to process commands.
        /// </summary>
        /// <param name="nodes">
        /// List of DMGateway machines to use. 
        /// If DMGateway Vip is enabled and dip fallback is disabled, this list should be empty.
        /// </param>
        /// <param name="dmGatewayVip">DMGateway Vip. If DMGateway Vip is disabled, this should be empty string.</param>
        /// <param name="sendTimeout">Timeout in milliseconds to send a command.</param>
        /// <param name="recvTimeout">Timeout in milliseconds to receive the response.</param>
        /// <param name="handler">Pointer to a DMClient::Handler object. When calls are made asynchronously,
        /// <param name="apContainerName">AP Container machine name</param>
        /// this callback will be invoked to indicate completion of the asynchronous request.</param>
        /// <returns>True if the intialization was successful, false otherwise.</returns>
        bool Init(
            _In_ const std::vector<std::string> &dmGatewayMachines,
            _In_ const std::string &dmGatewayVip,
            _In_ const UInt32 sendTimeout,
            _In_ const UInt32 recvTimeout,
            _In_ DMClient::Handler *handler,
            _In_ std::string apContainerName);

        /// <summary>
        /// Sends a command to the DM.
        /// </summary>
        /// <param name="readOnly">if true, it enables fast read.</param>
        /// <param name="cmd">command to be sent.</param>
        /// <param name="upload">additional data to be sent with the command.</param>
        /// <param name="resp">response from the DM.</param>
        /// <param name="ctx">command context.</param>
        /// <param name="traceId">A unique string that can be associated with the request, which helps in correlating logs.</param>
        bool Send(std::shared_ptr<DMClientImpl> thisPtr, bool readOnly, _In_z_ const char *cmd, CsMemoryBuffer *upload, DMResponse *resp, void *ctx, _In_opt_z_ const char* traceId);

        /// <summary>
        /// Sends a command to either ApManagementProxy, or DMGateway (depends on the derrived type of this DMClientImpl instance).
        /// </summary>
        /// <param name="thisPtr">The shared_ptr object that is storing the pointer to the current DMClientImpl instance.</param>
        /// <param name="req">Request encapsulating the command to be sent plus any additional payload.</param>
        /// <param name="resp">Response from ApManagementProxy / DMGateway.</param>
        /// <param name="ctx">Command context. Used to indicate whether the command should be executed asynchronously or not.</param>
        /// <returns>true if the command was successfully sent. Else false.</returns>
        virtual bool Send(std::shared_ptr<DMClientImpl> thisPtr, DMRequest* req, DMResponse *resp, void *ctx) = 0;

        /// <summary>
        /// Sets the maximum length for a message in bytes.
        /// </summary>
        /// <param name="maxMessageLen">Max message length.</param>
        virtual void SetMaxMessageLength(UInt32 /*maxMessageLen*/)
        {
        }

        /// <summary>
        /// Keep alive the connection to the server.
        /// </summary>
        /// <param name="flag">if set to true, the idle connections to the server should not be closed.</param>
        virtual void SetKeepAlive(bool flag)
        {
            m_keepAlive = flag;
        }

    protected:

        /// <summary>
        /// After DMClientImpl performs the common initialization tasks (e.g., setting the timeouts,
        /// the nodes, etc); this method is invoked so that derived classes can perform initialization
        /// tasks specific to their implementation.
        /// </summary>
        virtual bool OnInit() = 0;

        /// <summary>
        /// Represents list of DMGateway machines to send requests.
        /// </summary>
        std::set<std::string> m_dmGatewayMachines;

        /// <summary>
        /// Represents DMGateway vip to send requests.
        /// </summary>
        std::string m_dmGatewayVip;

        /// <summary>
        /// Timeout in milliseconds to send a command.
        /// </summary>
        UInt32 m_sendTimeout;

        /// <summary>
        /// Timeout in milliseconds to receive a response.
        /// </summary>
        UInt32 m_recvTimeout;

        DMClient::Handler *m_callback;
        bool m_keepAlive;

        std::string m_appName;
        std::string m_userName;
        std::string m_machineName;
        std::string m_environmentName;
        std::string m_clusterName;

        /// <summary>
        /// Ap container machine name.
        /// </summary>
        std::string m_apContainerName;

    private:

        /// <summary>
        /// Populates the caller data (i.e., data about the consumer of this instance of DMClient).
        /// This includes information such as the service name, the user name (under which this process
        /// is running), the current machine name, environment name, cluster name.
        /// All this will be sent to the DMGateway during command execution. This information will be used
        /// for diagnostics.
        /// </summary>
        void PopulateCallerMetadata();
    };
};

