#pragma once
#include "ap_pal.h"

#include <memory>

namespace dmclient
{
    /// <summary>
    /// This namespace encapsulates DMClient side perf counters used to monitor dmclient command request rates and responses.
    /// </summary>
    namespace PerfCounters
    {
        /// <summary>
        /// Thread safe initialize of DMClientPerfCounters class.
        /// Should be invoked before calling any other method in the class.
        /// </summary>
        /// <returns>True, on successful initialization. False, otherwise.</returns>
        bool Init();

        /// <summary>
        /// Invoked when a request is sent to DMGateway.
        /// Increments the counter tracking outgoing request rate to DMGateway.
        /// </summary>
        void OnRequestSent();

        /// <summary>
        /// Invoked when there is a connection failure for a request sent to DMGateway.
        /// Increments the counter tracking DMGateway connection failure rate.
        /// </summary>
        void OnConnectionFailure();

        /// <summary>
        /// Invoked after receiving the response on first successful connection for a request sent to DMGateway dip.
        /// Updates the counters tracking response latencies from DMGateway dip.
        /// </summary>
        /// <param name="requestAttempts">Number of times the request was sent to DMGateway till successful connection</param>
        /// <param name="requestStartTime">Time when the first request was sent to DMGateway.</param>
        /// <param name="successConnectionRequestStartTime">Time when sucessfull connection request was sent to DMGateway.</param>
        void OnSuccessfulConnection(_In_ const UInt32 requestAttempts, _In_ const LARGE_INTEGER& requestStartTime, _In_ const LARGE_INTEGER& successConnectionRequestStartTime);

        /// <summary>
        /// Calculates the time elapsed from input start time till current time.
        /// </summary>
        /// <param name="startTime">Input start time from which the latency should be calculated.</param>
        /// <returns>Latency in milli seconds.</returns>
        UInt64 GetCurrentLatencyInMilliSec(_In_ const LARGE_INTEGER& startTime);

        /// <summary>
        /// Calculates the time elapsed from input start time till input end time.
        /// </summary>
        /// <param name="startTime">Input start time from which the latency should be calculated.</param>
        /// <param name="endTime">Input end time till which the latency should be calculated.</param>
        /// <returns>Latency in milli seconds.</returns>
        UInt64 GetTimeDiffInMilliSec(_In_ const LARGE_INTEGER& startTime, _In_ const LARGE_INTEGER& endTime);
    }
}