#pragma once
#include "ap_pal.h"

#include "basic_types.h"
#include "Configuration.h"

#include <memory>

namespace apsdk
{
    class DMClient;
}

namespace dmclienttest
{
    class DMClientSettingsTest;
}

namespace dmclient
{
    class DMClientHttpImpl;
    class DMClientTestHelper;

    /// <summary>
    /// This enumeration captures various settings related to using DMGateway Vip.
    /// </summary>
    enum class UseDmGatewayVip
    {
        /// <summary>
        /// Do not use DMGateway vip. Use DMGateway machines directly to send requests.
        /// </summary>
        False,

        /// <summary>
        /// Use DMGateway vip, without falling back to DMGateway machines on vip connection failure.
        /// </summary>
        True,

        /// <summary>
        /// Use DMGateway vip, additionally fallback to DMGateway machines on vip connection failure.
        /// </summary>
        DipFallback,

        /// <summary>
        /// Use DMGateway vip configuration is not defined.
        /// </summary>
        NotDefined
    };

    /// <summary>
    /// This class is used to simplify the code that reads a configuration value from the config file and then updates
    /// the DMClientSettings object if the previous value was different.
    /// </summary>
    template<class T>
    class SettingsHelper
    {
    public:
        /// <summary>
        /// Updates a setting with a new value read from config. If a value wasn't read, then we use NULL (i.e.
        /// we don't use any default value).
        /// </summary>
        /// <param name="settingValue">Stores the setting value.</param>
        /// <param name="config">The configuration.</param>
        /// <param name="sectionName">The section name.</param>
        /// <param name="settingName">The setting name.</param>
        /// <param name="modified">This is set to true if the setting value has been updated. If the setting value is not updated,
        /// this (the value of "modified") is not touched.</param>
        /// <returns>True if the setting was found in the configuration file, else false.</returns>
        static bool UpdateSetting(
            _Inout_ std::unique_ptr<T>& settingValue,
            _In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config,
            _In_z_ const char* sectionName,
            _In_z_ const char* settingName,
            _Inout_ bool& modified);

        /// <summary>
        /// Updates a setting with a new value read from config. If a value wasn't read, then we use a default value.
        /// </summary>
        /// <param name="settingValue">Stores the setting value.</param>
        /// <param name="defaultValue">The default value, if the value wasn't present in the config.</param>
        /// <param name="config">The configuration.</param>
        /// <param name="sectionName">The section name.</param>
        /// <param name="settingName">The setting name.</param>
        /// <param name="modified">This is set to true if the setting value has been updated. If the setting value is not updated,
        /// this (the value of "modified") is not touched.</param>
        /// <returns>True if the setting was found in the configuration file, else false.</returns>
        static bool UpdateSetting(
            _Inout_ T& settingValue,
            _In_ const T& defaultValue,
            _In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config,
            _In_z_ const char* sectionName,
            _In_z_ const char* settingName,
            _Inout_ bool& modified);

    private:
        /// <summary>
        /// Reads a value for a setting from config.
        /// If the setting is not found, the default value is returned.
        /// </summary>
        static T GetSettingValueFromConfig(
            _In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config,
            _In_z_ const char* sectionName,
            _In_z_ const char* settingName,
            _In_ const T& defaultValue);
    };

    /// <summary>
    /// This class contains settings that are used by the DMClient library. The settings are used
    /// to determine whether DMClient should connect to ApManagementProxy or DMGateway; what is the
    /// DMGateway endpoint URL / port; whether DMGateway uses HTTPS or HTTP, etc.
    /// This class also acts as a IConfigurationListener, receiving notifications when configuration
    /// files have changed. Instances of this class are used to register listeners for the cluster
    /// shared.ini, environment environment.ini, and service default configuration files.
    /// </summary>
    class DMClientSettings : public apsdk::configuration::IConfigurationListener
    {
        friend class apsdk::DMClient;
        friend class dmclient::DMClientHttpImpl;
        friend class dmclient::DMClientTestHelper;
        friend class dmclienttest::DMClientSettingsTest;

    public:
        typedef void(*OnSettingsChanged)(void *cookie);

        /// <summary>
        /// Destructor.
        /// </summary>
        ~DMClientSettings();

        /// <summary>
        /// Creates an instance of the DMClientSettings object from the settings specified in the
        /// cluster shared.ini configuration.
        /// </summary>
        /// <param name="onSettingsChangedFunctionPointer">A function pointer that will be invoked when the cluster shared.ini
        /// file is modified.</param>
        /// <param name="cookie">A user provided object that will be passed back when onSettingsChangedFunctionPointer is invoked.
        /// The user code can use this object for storing state / correlation / etc.</param>
        /// <returns>A DMClientSettings object corresponding to the cluster settings.</returns>
        static std::shared_ptr<DMClientSettings> CreateFromClusterSettings(OnSettingsChanged onSettingsChangedFunctionPointer, _In_opt_ void* cookie);

        /// <summary>
        /// Creates an instance of the DMClientSettings object from the settings specified in the
        /// environment environment.ini configuration.
        /// </summary>
        /// <param name="onSettingsChangedFunctionPointer">A function pointer that will be invoked when the environment environment.ini
        /// file is modified.</param>
        /// <param name="cookie">A user provided object that will be passed back when onSettingsChangedFunctionPointer is invoked.
        /// The user code can use this object for storing state / correlation / etc.</param>
        /// <returns>A DMClientSettings object corresponding to the environment settings.</returns>
        static std::shared_ptr<DMClientSettings> CreateFromEnvironmentSettings(OnSettingsChanged onSettingsChangedFunctionPointer, _In_opt_ void* cookie);

        /// <summary>
        /// Creates an instance of the DMClientSettings object from the settings specified in the
        /// service default configuration.
        /// </summary>
        /// <param name="onSettingsChangedFunctionPointer">A function pointer that will be invoked when the service default configuration
        /// file is modified.</param>
        /// <param name="cookie">A user provided object that will be passed back when onSettingsChangedFunctionPointer is invoked.
        /// The user code can use this object for storing state / correlation / etc.</param>
        /// <returns>A DMClientSettings object corresponding to the service (application) settings.</returns>
        static std::shared_ptr<DMClientSettings> CreateFromApplicationSettings(OnSettingsChanged onSettingsChangedFunctionPointer, _In_opt_ void* cookie);

        /// <summary>
        /// Gets the UseDmGatewayVip enum value of the input string.
        /// </summary>
        /// <param name="useDmGatewayVipStr">Input string value of the enumeration UseDmGatewayVip.</param>
        /// <returns>UseDmGatewayVip, enum value of the input string.</returns>
        static UseDmGatewayVip ParseUseDmGatewayVipSetting(_In_ const std::string& useDmGatewayVipStr);

        /// <summary>
        /// Gets UseDmGatewayVip setting based on input config settings.
        /// Application settings override environment settings. Environment settings override cluster settings.
        /// </summary>
        /// <param name = "applicationSettings">application level settings</param>
        /// <param name = "environmentSettings">environment level settings</param>
        /// <param name = "clusterSettings">cluster level settings</param>
        static UseDmGatewayVip GetUseDMGatewayVip(
            _In_ const std::shared_ptr<dmclient::DMClientSettings> applicationSettings,
            _In_ const std::shared_ptr<dmclient::DMClientSettings> environmentSettings,
            _In_ const std::shared_ptr<dmclient::DMClientSettings> clusterSettings);

        /// <summary>
        /// Gets UseDmGatewayVip setting.
        /// </summary>
        /// <returns>UseDmGatewayVip, that indicates whether to use vip or not.</returns>
        UseDmGatewayVip GetUseDmGatewayVipSetting();

        /// <summary>
        /// Populates the information about the DMGateway endpoint to which DMClient should connect to.
        /// </summary>
        /// <param name="resourceUrl">Populated with the HTTP resource URL (the path of the resource on the server), corresponding
        /// to DMGateway.</param>
        /// <param name="port">Populated with the port number of the endpoint where DMGateway is listening.</param>
        /// <param name="useHttps">Populated with true/false to indicate if an SSL connection should be established.</param>
        /// <param name="serverCertificateAcl">Populated with the ACL that can be used to verify the DMGateway SSL certificate, if an SSL connection is used.</param>
        void GetDmGatewayEndpointSettings(_Out_ std::string& resourceUrl, _Out_ unsigned short& port, _Out_ bool& useHttps, _Out_ std::string& serverCertificateAcl);

        /// <summary>
        /// Gets the DMGateway vip of the current cluster from the cluster settings.
        /// Should be invoked only on cluster settings instance.
        /// </summary>
        /// <returns>
        /// DMGateway vip of the current cluster.
        /// Returns empty string if vip is not defined in cluster shared.ini
        /// </returns>
        std::string GetDmGatewayVip();

        /// <summary>
        /// Unregisters the handler that gets invoked whenever settings change.
        /// </summary>
        void UnRegisterSettingsChangeHandler();

        /// <summary>
        /// Invoked when the configuration file has changed. The implementation here
        /// can decide how to handle the changes.
        /// Here (for DMClient), we just reload the DMClient settings from the appropriate
        /// configuration file (shared.ini, environment.ini, or default configuration of 
        /// service).
        /// </summary>
        /// <param name="name">the configuration name used to register the listener.</param>
        /// <param name="pathname">the fully qualified pathname of the configuration</param>
        /// <param name="cookie">the original cookie used to register the listener.</param>
        /// <param name="configuration">the new (changed) configuration instance.</param>
        virtual void ConfigurationChanged(
            const char* name,
            const char* pathname,
            void* cookie,
            apsdk::Ptr<const apsdk::configuration::IConfiguration> configuration) override;

    private:

        /// <summary>
        /// This enumeration captures whether settings are associated with the cluster shared.ini file,
        /// or the environment environment.ini file, or the application default configuration file.
        /// </summary>
        enum SettingsLevel
        {
            /// <summary>
            /// Settings are associated with the cluster shared.ini.
            /// </summary>
            SettingsLevel_cluster,

            /// <summary>
            /// Settings are associated with the environment environment.ini.
            /// </summary>
            SettingsLevel_environment,

            /// <summary>
            /// Settings are associated with the application default configuration file.
            /// </summary>
            SettingsLevel_application
        };

        /// <summary>
        /// Array having string values for each of the enumeration of DMGatewayUseVip in the order of declaration.
        /// </summary>
        static const char *c_useDmGatewayVipString[];

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="settingsLevel">Indicates whether these settings represent cluster / environment / service settings.</param>
        /// <param name="onSettingsChangedFunctionPointer">A function pointer that will be invoked when the underlying configuration
        /// file is modified.</param>
        /// <param name="cookie">A user provided object that will be passed back when onSettingsChangedFunctionPointer is invoked.
        /// The user code can use this object for storing state / correlation / etc.</param>
        DMClientSettings(SettingsLevel settingsLevel, OnSettingsChanged onSettingsChangedFunctionPointer, _In_opt_ void* cookie);

        /// <summary>
        /// Constructor. Used by test code.
        /// </summary>
        DMClientSettings();

        /// <summary>
        /// Loads the DMClient settings from the cluster shared.ini file.
        /// </summary>
        /// <param name="config">The configuration corresponding to shared.ini.</param>
        /// <returns>True if the settings were modified, False if not.</returns>
        bool LoadClusterSettings(_In_ apsdk::Ptr<const apsdk::configuration::IConfiguration>& config);

        /// <summary>
        /// Loads the DMClient settings from the environment environment.ini file.
        /// </summary>
        /// <param name="config">The configuration corresponding to environment.ini.</param>
        /// <returns>True if the settings were modified, False if not.</returns>
        bool LoadEnvironmentSettings(_In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config);

        /// <summary>
        /// Loads the DMClient settings from the application default configuration file.
        /// </summary>
        /// <param name="config">The application's default configuration.</param>
        /// <returns>True if the settings were modified, False if not.</returns>
        bool LoadApplicationSettings(_In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config);

        /// <summary>
        /// Loads the common DMClient settings from the passed in configuration file. The settings that
        /// are loaded here are those which are valid at all levels - cluster, environment, and application.
        /// </summary>
        /// <param name="config">The configuration corresponding to either shared.ini, environment.ini, or the
        /// application default configuration file.</param>        
        /// <returns>True if the settings were modified, False if not.</returns>
        bool LoadCommonSettings(_In_ const apsdk::Ptr<const apsdk::configuration::IConfiguration>& config);

        /// <summary>
        /// The name of the section in the configuration file where the DMClient settings are read from.
        /// </summary>
        static const char* c_sectionName;

        /// <summary>
        /// The name of the setting which determines whether the DMClient should only use the Machine Function certificate
        /// for authentication with DMGateway.
        /// </summary>
        static const char* c_onlyUseMachineCertificateKey;

        /// <summary>
        /// The name of the setting which determines whether the "OnlyUseMachineCertification" setting value can be overridden
        /// at the environment / service level.
        /// </summary>
        static const char* c_onlyUseMachineCertificateOverridableKey;

        /// <summary>
        /// The name of the setting whose value indicates the port number at which the DMGateway is listening for requests.
        /// </summary>
        static const char* c_dmGatewayPortKey;

        /// <summary>
        /// The name of the setting whose value indicates the HTTP resource URL to which the DMGateway requests need to
        /// be sent.
        /// </summary>
        static const char* c_dmGatewayResourceUrlKey;

        /// <summary>
        /// The name of the setting whose value indicates whether an HTTPS connection (as opposed to HTTP) needs to be 
        /// used when connecting to the DMGateway.
        /// </summary>
        static const char* c_dmGatewayUseHttpsKey;

        /// <summary>
        /// The name of the setting whose value contains the AP PKI ACL which will be used to validate the service
        /// certificate, to ensure that we are indeed connecting to a DMGateway machine and no other.
        /// </summary>
        static const char* c_dmGatewaySslCertificateAclKey;

        /// <summary>
        /// The name of the setting whose value indicates whether to use DMGateway vip or DMGateway machines
        /// for sending requests.
        /// </summary>
        static const char* c_useDmGatewayVipKey;

        /// <summary>
        /// The name of the setting which gives DMGateway vip.
        /// </summary>
        static const char* c_dmGatewayVipKey;

        /// <summary>
        /// The name of the setting whose value contains the maximum number of connections that should be opened
        /// to a single endpoint.
        /// </summary>
        static const char* c_maxConnectionsPerServerKey;

        /// <summary>
        /// Controls the number of times to retry a command in case of a failure.
        /// </summary>
        static const char* c_numRetriesKey;

        /// <summary>
        /// The name of the setting that controls whether we should continue retrying even if the 
        /// initial timeout specified by the user has expired.
        /// </summary>
        static const char* c_ignoreTimeoutDuringRetryKey;

        /// <summary>
        /// The name of the setting that controls the timeout for name resolution.
        /// </summary>
        static const char* c_resolveTimeoutMillisecondsKey;

        /// <summary>
        /// The name of the setting that controls the timeout for establishing the connection to the server.
        /// </summary>
        static const char* c_connectTimeoutMillisecondsKey;

        /// <summary>
        /// The name of the setting that controls the timeout for sending requests.
        /// </summary>
        static const char* c_sendTimeoutMillisecondsKey;

        /// <summary>
        /// The name of the setting that controls the timeout for receiving requests.
        /// </summary>
        static const char* c_receiveTimeoutMillisecondsKey;

        /// <summary>
        /// The default value to use for the "UseMachineCertificateOverridable" setting, if the value is not found in the
        /// configuration file.
        /// </summary>
        static const bool c_onlyUseMachineCertificateOverridableDefault;

        /// <summary>
        /// The default value to use for the "DmGatewayPort" setting, if the value is not found in the
        /// configuration file.
        /// </summary>
        static const unsigned short c_dmGatewayPortDefault;

        /// <summary>
        /// The default value to use for the "DmGatewayResourceUrl" setting, if the value is not found in the
        /// configuration file.
        /// </summary>
        static const char* c_dmGatewayResourceUrlDefault;

        /// <summary>
        /// The default value to use for the "DmGatewayUseHttps" setting, if the value is not found in the
        /// configuration file.
        /// </summary>
        static const bool c_dmGatewayUseHttpsDefault;

        /// <summary>
        /// The default value to use for the "DmGatewaySslCertificateAcl" setting, if the value is not found
        /// in the configuration file.
        /// </summary>
        static const char* c_dmGatewaySslCertificateAclDefault;

        /// <summary>
        /// The default value to use for the "MaxConnectionsPerServer" setting, if the value is not found
        /// in the configuration file.
        /// </summary>
        static const unsigned long c_maxConnectionsPerServerDefault;

        /// <summary>
        /// The default value to use for the "NumRetries" setting, if the value is not found
        /// in the configuration file.
        /// </summary>
        static const int c_numRetriesDefault;

        /// <summary>
        /// The default value to use for the "IgnoreTimeoutDuringRetry" setting, if the value
        /// is not found in the configuration file.
        /// </summary>
        static const bool c_ignoreTimeoutDuringRetryDefault;

        /// <summary>
        /// The default value for the timeout for name resolution.
        /// </summary>
        static const unsigned int c_resolveTimeoutMillisecondsDefault;

        /// <summary>
        /// The default value for the timeout for establishing the connection to the server.
        /// </summary>
        static const unsigned int c_connectTimeoutMillisecondsDefault;

        /// <summary>
        /// The listener handler returned by the Configuration::RegisterListener API. This corresponds
        /// to the listener registration for the underlying configuration file for this DMClientSettings object.
        /// </summary>
        void* m_changeListenerHandle;

        /// <summary>
        /// Synchronizes access to the settings, across "Read" of the settings, and the "Write" to the settings
        /// when the change listener gets triggered.
        /// </summary>
        CRITSEC m_lock;

        /// <summary>
        /// The HTTP resource URL (the fragment after the host name in the HTTP URL) for the DMGateway. The
        /// DMClient specifies this to the HttpRequest instance when connecting to DMGateway.
        /// This setting will be set only if this DMClientSettings object is representing the cluster settings.
        /// </summary>
        std::string m_dmGatewayResourceUrl;

        /// <summary>
        /// The port at which the DMGateway is listening for incoming requests.
        /// This setting will be set only if this DMClientSettings object is representing the cluster settings.
        /// </summary>
        unsigned short m_dmGatewayPort;

        /// <summary>
        /// Indicates whether an HTTPS connection needs to be established to DMGateway, or HTTP.
        /// This setting will be set only if this DMClientSettings object is representing the cluster settings.
        /// </summary>
        bool m_dmGatewayUsesHttps;

        /// <summary>
        /// The AP PKI ACL that will be used to validate the server SSL certificate given by DMGateway,
        /// to ensure that the DMClientHttpImpl is indeed connecting to the DMGateway and not some other
        /// service.
        /// </summary>
        std::string m_dmGatewaySslCertificateAcl;

        /// <summary>
        /// Determines whether to use vip (with/without dip fallback) or to use DMGateway machines for sending requests.
        /// </summary>
        UseDmGatewayVip m_useDmGatewayVip;

        /// <summary>
        /// The DMGateway vip to be used to send requests.
        /// Defined in cluster level settings.
        /// </summary>
        std::string m_dmGatewayVip;

        /// <summary>
        /// Indicates whether only the machine function certificate needs to be used by DMClient
        /// when connecting to DMGateway (i.e., we explicitly should use the machine function certificate
        /// and not the user-context certificate). 
        /// The reason we have this as a setting is for the initial stages when rolling out this new DMClient / DMGateway;
        /// if a customer is running his/her service under a user account, there won't be a user-certificate available
        /// (today, AP PKI infra only provisions a user-certificate when a user logs in via remote desktop). And thus,
        /// the DMGateway will fail the DMClient request.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then (in the current implementation) the behavior is
        /// as if this is true.
        /// </summary>
        std::unique_ptr<bool> m_onlyUseMachineCertificate;

        /// <summary>
        /// Controls the maximum number of connections opened to a single DMGateway endpoint by DMClient.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<unsigned long> m_maxConnectionsPerServer;

        /// <summary>
        /// Controls the number of times DMClient will retry on failures, after the first execution
        /// of a command has failed.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<int> m_numRetries;

        /// <summary>
        /// The timeout for name resolution.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<unsigned int> m_resolveTimeoutMilliseconds;

        /// <summary>
        /// The timeout for establishing the connection to the server.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<unsigned int> m_connectTimeoutMilliseconds;

        /// <summary>
        /// The timeout for sending requests.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<unsigned int> m_sendTimeoutMilliseconds;

        /// <summary>
        /// The timeout for receiving requests.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<unsigned int> m_receiveTimeoutMilliseconds;

        /// <summary>
        /// Controls whether the timeout should be ignored if we are retrying a command.
        /// If NULL, it means this wasn't configured at this level.
        /// If not configured at any level, then the default value is used.
        /// </summary>
        std::unique_ptr<bool> m_ignoreTimeoutDuringRetry;

        /// <summary>
        /// Indicates whether the configuration for OnlyUseMachineCertificate is overridable at the environment/service
        /// level.
        /// This setting will be set only if this DMClientSettings object is representing the cluster settings.
        /// </summary>
        bool m_onlyUseMachineCertificateOverridable;

        /// <summary>
        /// Indicates whether this DMClientSettings object has been constructed for reading settings from the
        /// cluster shared.ini, or the environment environment.ini, or the application default configuration.
        /// </summary>
        SettingsLevel m_settingsLevel;

        /// <summary>
        /// A function pointer that will be invoked if the underlying configuration file
        /// for this settings object is modified.
        /// This value is provided in the constructor.
        /// </summary>
        OnSettingsChanged m_onSettingsChangedFunctionPointer;

        /// <summary>
        /// This represents the "state" that will be used when invoking
        /// m_onSettingsChangedFunctionPointer.
        /// This value is provided in the constructor.
        /// </summary>
        void* m_onSettingsChangedFunctionPointerCookie;
    };
};
